"""Front models

    The object types in front, modeled with simple methods for accessing and manipulating the data.

    Example: 

        from models.client import FrontAppClient
        from globals import TOKEN
        from pprint import pprint

        # create an authorized client
        front = FrontAppClient(TOKEN) 
        
        # returns the list of Inbox objects
        authorized_inboxes = front.inboxes 
        
        # returns the first Inbox in the list of Inbox objects
        two_week_inbox = front.inboxes[0] 
        
        # returns the first Conversation in the list of Conversation objects
        meter_location_email = two_week_inbox.conversations[0]         
        
        # returns all the Comment objects from a specific conversation 
        comments = meter_location_email.get_comments() 

"""

# imports all libraries 
import os
from types import new_class
import requests
import sys
import json
import pprint
from typing import Any, Dict, List, Optional, Union
import sys
from datetime import datetime, timedelta
import urllib.parse as parse

# frontapp
from frontapp.globals import LIBRARY, API_URL, TOKEN


class BaseModel:
    """basic model class which parses data and setts attributes"""

    def __init__(self, data: dict, client: Any = None, **kwargs: Any) -> None:
        """constructor"""
        self.id = None
        self._data = {**data, **kwargs}
        self._json = self._jsond(data)
        self._client = client

        for key in self._data:
            setattr(self, key, self._data[key])

    def _jsond(self, json_data: dict) -> str:
        """json dumps"""
        return json.dumps(json_data)

    def _jsonl(self, dictionary: str) -> dict:
        """json loads"""
        return json.loads(dictionary)


class Comment(BaseModel):
    """class model for Comments"""

    def __init__(self, data, **kwargs):
        super(Comment, self).__init__(data, **kwargs)

    def __repr__(self) -> str:
        return (
            f"<{LIBRARY}.Comment[{self.id:}] '{self.author['first_name']}: {self.body}'"
        )


class Conversation(BaseModel):
    """Class model for Conversations"""

    def __repr__(self):
        return f"<{LIBRARY}.Conversation[{self.id}] '{self.subject}'>"

    def comment(
        self,
        message: str,
        author_id: str = "alt:email:",
    ) -> Comment:
        """posts comments into conversations"""
        body = {"author_id": author_id, "body": message}
        response = self._client.post(f"/conversations/{self.id}/comments", json=body)
        if "_error" in response:
            raise ConnectionError(json.dumps(response))

        return Comment(response, client=self._client)

    def get_comments(self) -> Comment:
        """Get the comments in the conversation"""
        comments_data = self._client.get(f"/conversations/{self.id}/comments")
        return [Comment(x, client=self._client) for x in comments_data["_results"]]

    def filter_comments(self, text: str):
        """Will look for certain comments"""
        url = API_URL + "conversations/" + self.conversation_id + "/comments"
        headers = {
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
            "Accept": "application/json",
        }
        response = requests.request("GET", url, headers=headers)
        text_repo = response.text
        commands_pre = text_repo.count(text)
        print(commands_pre)
        return commands_pre

    # this is the def that could be implemented that would get emails only 2 weeks ago. problem: each api requests only gets up to 100 and there are more than 100 emails in 2 weeks
    def get_emails(self, time=50, limit="50"):
        """gets conversations id's and subjects based of parameters given"""
        dt = datetime.timestamp(datetime.now() - timedelta(hours=time))
        dt = str(dt)
        dt = dt[:-3]
        url = API_URL + "conversations/search/after%3A" + dt + "?limit=" + limit
        headers = {
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
            "Accept": "application/json",
        }

        response = requests.request("GET", url, headers=headers)
        data = response.json()
        entries = []
        for entry in data["_results"]:
            id = entry["id"]
            subject = entry["subject"]
            entries.append(id + ":" + subject)

        return entries

    def reply(self, client: str, sender: str, subject: str, id: str, body: str):
        """replies to emails"""
        url = (
            API_URL + "conversations/" + self.conversation_id + "/messages"
        )  # conversation id
        print(url)
        payload = {
            "to": [client],
            "options": {"archive": True},
            "sender_name": sender,
            "subject": subject,
            "author_id": id,
            "body": body,
        }
        headers = {
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        response = requests.request("POST", url, json=payload, headers=headers)
        print(response.text)


class Channel(BaseModel):
    """Class model for Channels"""

    def __repr__(self) -> str:
        """Represents"""
        return f"<{LIBRARY}.Channel[{self.id}] '{self.name}'>"

    def create_message(
        self,
        to: str,
        sender_name: str,
        subject: str,
        author_id: str,
        body: str,
    ):
        url = (
            API_URL + "channels/" + self.channel_id + "/messages"
        )  # sending channel id
        payload = {
            "to": to,
            "options": {"archive": True},
            "sender_name": sender_name,
            "author_id": author_id,
            "subject": subject,
            "body": body,
        }
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("POST", url, json=payload, headers=headers)
        print(response.text)


class Inbox(BaseModel):
    """Inbox object"""

    @property
    def channels(self) -> List[Channel]:
        """get the channels this inbox has access to"""
        channel_data = self._client.get(f"inboxes/{self.id}/channels")
        if "_error" in channel_data:
            raise ConnectionError(json.dumps(channel_data))

        if not isinstance(channel_data, dict):
            raise Exception("Invalid response while looking up inboxes.")

        self._channels = [
            Channel(x, client=self._client) for x in channel_data["_results"]
        ]
        return self._channels

    # note: this is probably impractical as a property - just keeping it here for testing. too much property nesting will end up with O(n^2) time complexity which would just take forever
    # idea: could make this api request limited to only the day's conversations, etc
    @property
    def conversations(self) -> List[Conversation]:
        """get the conversations in this inbox"""
        conversation_data = self._client.get(f"inboxes/{self.id}/conversations")
        if "_error" in conversation_data:
            raise ConnectionError(json.dumps(conversation_data))

        if not isinstance(conversation_data, dict):
            raise Exception("Invalid response while looking up inboxes.")

        self._conversations = [
            Conversation(x, client=self._client) for x in conversation_data["_results"]
        ]

        # return conversation_data
        return self._conversations

    def __repr__(self):
        return f"<{LIBRARY}.Inbox[{self.id}] '{self.name}'>"

    def get_specific_message(self):
        """list all details about messages"""
        url = API_URL + "messages/" + self.message_id  # message id
        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("GET", url, headers=headers)
        print(response.text)

    def get_teams(self) -> List[Conversation]:
        url = API_URL + "teams"

        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }

        response = requests.request("GET", url, headers=headers)

        return response.text


class Tag(BaseModel):

    def __repr__(self):
        return f"<{LIBRARY}.Tag[{self.id}] '{self.name}'>"




    def update_name(self, name: str, visibility: bool, color="red"):
        """updates name, highlight, visibility of team tags."""
        url = API_URL + "tags/" + self.tag_id
        payload = {
            "is_visible_in_conversation_lists": visibility,
            "highlight": color,
            "name": name,
        }
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("PATCH", url, json=payload, headers=headers)
        print(response.text)

    def assign_tag(self, conversation_id: str):
        """assigns a tag to email/conversation"""
        url = API_URL + "conversations/" + conversation_id + "/tags"
        payload = {"tag_ids": [self.tag_id]}
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("POST", url, json=payload, headers=headers)
        print(response.text)

    def get_tagged_conversatons(self):
        """pulls all conversations/email data from specific tags"""
        url = API_URL + "tags/" + self.tag_id + "/conversations"
        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("GET", url, headers=headers)
        parsed = json.loads(response.text)
        print(parsed)


class Team(BaseModel):
    def __repr__(self):
        return f"<{LIBRARY}.Conversation[{self.id}] '{self.subject}'>"

    def get_teams(self) -> List[Conversation]:
        url = API_URL + "teams"

        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }

        response = requests.request("GET", url, headers=headers)

        return response.text

    pass


# for personal use rn
class Retrive:
    def list_tag(self):
        """List all details about tags"""
        url = API_URL + "tags"
        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("GET", url, headers=headers)
        print(response.text)
