"""
# Base client model to create and use http endpoints.

        TODO: break BaseClient as separate package for other uses. Consider making BaseClient an ABS class with abstractmethods.
"""

from re import L
import requests
import urllib.parse
from requests.models import Response
from frontapp.globals import __version__, API_URL, LIBRARY
from typing import Any, Dict, List, Optional, Union
from pprint import pprint
import json
import os

# frontapp
from frontapp.models import Inbox, Conversation, Channel, Tag


class BaseClient:
    """base client http wrapper"""

    def __init__(
        self,
        token: str,
        api_url: str,
        user_agent: str = f"{LIBRARY}/{__version__}",
        cache: bool = False,
        debug: bool = False,
    ) -> None:
        """creates a new client"""
        if not token:
            raise Exception("no token specified!")

        self.token = token
        self.api_url = api_url
        self.version = __version__
        self.cache = cache
        self.debug = debug
        self.user_agent = user_agent

    @property
    def headers(self) -> dict:
        """forms the headers required for the API calls"""
        return {
            "Accept": "application/json",
            "AcceptEncoding": "gzip, deflate",
            "Authorization": self.token,
            "User-Agent": self.user_agent,
        }

    def _log(self, *args: Any) -> None:
        """logging method"""
        if not self.debug:
            return
        pprint(*args)

    def _req(self, path: str, method: str = "get", **kwargs: Any) -> Response:
        """requests wrapper"""
        full_path = urllib.parse.urljoin(self.api_url, path)
        self._log(f"[{method.upper()}]: {full_path}")
        request = requests.request(method, full_path, headers=self.headers, **kwargs)
        # TODO: raise exeption if rate limit is reached
        return request

    def get(
        self, path: str, raw: bool = False, **kwargs: Any
    ) -> Union[list, dict, Response]:
        """makes a get request to the API"""
        request = self._req(path, **kwargs)
        return request if raw else request.json()

    def post(
        self, path: str, raw: bool = False, **kwargs: Any
    ) -> Union[list, dict, Response]:
        """makes a post request to the API"""
        request = self._req(path, method="post", **kwargs)
        return request if raw else request.json()

    def put(
        self, path: str, raw: bool = False, **kwargs: Any
    ) -> Union[list, dict, Response]:
        """makes a put request to the API"""
        request = self._req(path, method="put", **kwargs)
        return request if raw else request.json()


class FrontAppClient(BaseClient):
    """base client http wrapper for FrontApp"""

    def __init__(self, token) -> None:
        super(FrontAppClient, self).__init__(
            token,
            api_url=API_URL,
            user_agent=f"{LIBRARY}/{__version__}",
        )

        # cache
        self._inboxes = None  # type: Optional[List[Inbox]]
        self._teams = None  # type: Optional[List[Team]]
        self._conversation = None  # type: Optional [List[Conversation]]
        self._tag = None  # type: Optional [List[tag]]
        self._channel = None  # type: Optional [List[Channel]]

    def __repr__(self):
        return f"<{LIBRARY}.FrontAppClient>"

    @property
    def inboxes(self) -> List[Inbox]:
        """get authorized inboxes"""
        if not self._inboxes or not self.cache:

            inbox_data = self.get("inboxes")
            if "_error" in inbox_data:
                raise ConnectionError(json.dumps(inbox_data))

            if not isinstance(inbox_data, dict):
                raise Exception("Invalid response while looking up inboxes.")

            self._inboxes = [Inbox(x, client=self) for x in inbox_data["_results"]]
            return self._inboxes

    @property
    def tag(self) -> List[Tag]:
        """get authorized tags"""
        if not self._tag or not self.cache:

            tag_data = self.get("tags")
            if "_error" in tag_data:
                raise ConnectionError(json.dumps(tag_data))

            if not isinstance(tag_data, dict):
                raise Exception("Invalid response while looking up tags.")

            self._tag = [Tag(x, client=self) for x in tag_data["_results"]]
            return self._tag

    @property
    def channel(self) -> List[Channel]:
        if not self._channel or not self.cache:

            channel_data = self.get("channels")
            if "_error" in channel_data:
                raise ConnectionError(json.dumps(channel_data))

            if not isinstance(channel_data, dict):
                raise Exception("Invalid response while looking up tags.")

            self._channel = [Channel(x, client=self) for x in channel_data["_results"]]
            return self._channel

    def get_team_id(self):
        """Get team information"""
        url = "https://api2.frontapp.com/teams"
        headers = {
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
            "Accept": "application/json",
        }
        response = requests.request("GET", url, headers=headers)
        print(response.text)
        return response.text

    def create_tag(name: str, team_id: str, visibility: bool, color="red"):
        """Create team tags"""
        url = "https://api2.frontapp.com/teams/" + team_id + "/tags"

        payload = {
            "is_visible_in_conversation_lists": visibility,
            "name": name,
            "highlight": color,
        }
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": "Bearer " + os.environ["AUTH_TOKEN"],
        }
        response = requests.request("POST", url, json=payload, headers=headers)
        print(response)
